"""BLS Provider Module."""

from openbb_bls.models.search import BlsSearchFetcher
from openbb_bls.models.series import BlsSeriesFetcher
from openbb_core.provider.abstract.provider import Provider

bls_provider = Provider(
    name="bls",
    website="https://www.bls.gov/developers/api_signature_v2.htm",
    description="The Bureau of Labor Statistics' (BLS) Public Data Application Programming Interface (API)"
    + " gives the public access to economic data from all BLS programs."
    + " It is the Bureau's hope that talented developers and programmers will use the BLS Public Data API to create"
    + " original, inventive applications with published BLS data.",
    credentials=["api_key"],
    fetcher_dict={
        "BlsSearch": BlsSearchFetcher,
        "BlsSeries": BlsSeriesFetcher,
    },
    repr_name="Bureau of Labor Statistics' (BLS) Public Data API",
    instructions="Sign up for a free API key here: https://data.bls.gov/registrationEngine/",
)



"""BLS Helpers."""

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from pandas import DataFrame


# We need to wrap this as a helper to accommodate requests for historical data
# greater than 20 years in length, or containing more than 50 symbols.
async def get_bls_timeseries(  # pylint: disable=too-many-branches,too-many-positional-arguments  # noqa: PLR0912
    api_key: str,
    series_ids: str | list[str],
    start_year: int | None = None,
    end_year: int | None = None,
    calculations: bool = True,
    catalog: bool = True,
    annual_average: bool = True,
    aspects: bool = False,
) -> dict:
    """Get BLS timeseries data. Max 50 symbols per request, and a 20 year historical window.

    Parameters
    ----------
    api_key : str
        BLS API key.
    series_ids : List[str]
        List of BLS series IDs. Max 50 symbols per request.
    start_year : Optional[int]
        Start year for the data. Max history per-request may span 20 years.
    end_year : Optional[int]
        End year for the data.  Max history per-request may span 20 years.
    calculations : bool
        Include calculations in the response, if available. Default is True.
    catalog : bool
        Include catalog information in the response. Default is True.
    annual_average : bool
        Include annual averages in the response, if available. Default is True.
    aspects : bool
        Include aspects in the response, if available. Default is False.

    Returns
    -------
    Dict
        Returns a dictionary with the following keys: data, metadata, messages.
    EmptyDataError
        If no data is found, an EmptyDataError is returned and not raised.
    """
    # pylint: disable=import-outside-toplevel
    import json  # noqa
    from warnings import warn
    from openbb_core.provider.utils.errors import EmptyDataError
    from openbb_core.provider.utils.helpers import amake_request

    symbols = series_ids.split(",") if isinstance(series_ids, str) else series_ids
    if len(symbols) > 50:
        warn(
            "Max 50 symbols per request. Truncating to 50 symbols."
            "Break the request into multiple queries to get more data."
        )
        symbols = symbols[:50]

    headers = {"Content-type": "application/json"}
    url = "https://api.bls.gov/publicAPI/v2/timeseries/data/"
    payload = json.dumps(
        {
            k: v
            for k, v in {
                "seriesid": symbols,
                "startyear": start_year,
                "endyear": end_year,
                "catalog": catalog,
                "calculations": calculations,
                "annualaverage": annual_average,
                "aspects": aspects,
                "registrationkey": api_key,
            }.items()
            if v
        }
    )
    res = await amake_request(url=url, method="POST", headers=headers, data=payload)
    results = res.get("Results", {}).get("series", [])  # type: ignore
    messages = res.get("message", [])  # type: ignore
    if messages:
        messages = [
            (
                f"The key provided by the User is invalid. {m.split(' provided by the User is invalid.')[1].strip()}"
                if m.startswith("The key:")
                else m
            )
            for m in messages
            if m
        ]

    metadata: dict = {}
    data: list = []
    for result in results:
        seriesID = result.get("seriesID")
        if not seriesID:
            continue
        catalog = result.get("catalog")
        if catalog:
            metadata.update({seriesID: catalog})
        _data = result.get("data", [])
        for _d in _data:
            new_d: dict = {}
            year = _d.get("year", "")
            month = _d.get("period", "").replace("M", "")
            if month.startswith("A") or month in ("S01", "Q01"):
                _date = year + "-01-01"
            elif month == "S02":
                _date = year + "-07-01"
            elif month in ("S03", "Q05"):
                _date = year + "-12-31"
                month = "13"
            elif month == "Q02":
                _date = year + "-04-01"
            elif month == "Q03":
                _date = year + "-07-01"
            elif month == "Q04":
                _date = year + "-10-01"
            else:
                _date = year + "-12-31" if month == "13" else year + "-" + month + "-01"
            new_d["symbol"] = seriesID
            title = metadata[seriesID].get("series_title") if catalog else None
            title = (
                title + (" (Annual Average)" if month == "13" else "")
                if title
                else None
            )
            if title:
                new_d["title"] = title
            new_d["date"] = _date
            value = _d.get("value")
            if value and value != "-":
                new_d["value"] = float(value)
            else:
                new_d["value"] = None
            new_d["latest"] = _d.get("latest") == "true"
            footnotes = _d.get("footnotes")
            if footnotes:
                new_d["footnotes"] = "; ".join(
                    [f.get("text") if isinstance(f, dict) else str(f) for f in footnotes if f]  # type: ignore
                )
                if not new_d.get("footnotes"):
                    new_d.pop("footnotes")

            calcs = _d.get("calculations")
            if calcs:
                changes = calcs.get("net_changes")
                pct_changes = calcs.get("pct_changes")
                if changes:
                    new_d["change_1M"] = (
                        float(changes.get("1")) if changes.get("1") else None
                    )
                    new_d["change_3M"] = (
                        float(changes.get("3")) if changes.get("3") else None
                    )
                    new_d["change_6M"] = (
                        float(changes.get("6")) if changes.get("6") else None
                    )
                    new_d["change_12M"] = (
                        float(changes.get("12")) if changes.get("12") else None
                    )

                if pct_changes:
                    new_d["change_percent_1M"] = (
                        float(pct_changes.get("1")) / 100
                        if pct_changes.get("1")
                        else None
                    )
                    new_d["change_percent_3M"] = (
                        float(pct_changes.get("3")) / 100
                        if pct_changes.get("3")
                        else None
                    )
                    new_d["change_percent_6M"] = (
                        float(pct_changes.get("6")) / 100
                        if pct_changes.get("6")
                        else None
                    )
                    new_d["change_percent_12M"] = (
                        float(pct_changes.get("12")) / 100
                        if pct_changes.get("12")
                        else None
                    )
            if aspects is True:
                # If there are aspects returned, we want to separate them from the main data.
                # We will store the aspects in the metadata dictionary.
                _aspects = _d.get("aspects")
                if _aspects:
                    for aspect in _aspects:
                        aspect.update(
                            {
                                "date": _date,
                                "footnotes": " ".join(
                                    [f.get("text", "") for f in footnotes] if footnotes else None  # type: ignore
                                ).strip(),
                            }
                        )
                    new_aspects = [
                        d for d in _aspects if d.get("value") and d.get("value") != "-"
                    ]
                    if new_aspects:
                        metadata[seriesID]["aspects"] = new_aspects
                    else:
                        messages.append(f"No Aspect Available for Series {seriesID}")

            if new_d:
                data.append(new_d)

    if not data:
        # Return EmptyDataError if no data is found instead of raising.
        # If we raise here, the API key can be exposed in the traceback.
        return EmptyDataError(f"No data found -> {messages}")  # type: ignore

    return {"data": data, "metadata": metadata, "messages": messages}


async def get_survey_asset(survey: str, asset: str) -> "DataFrame | None":
    """Get an asset in the FTP download folder of the two-letter survey code."""
    # pylint: disable=import-outside-toplevel
    from io import StringIO  # noqa
    from openbb_core.provider.utils.helpers import make_request
    from numpy import nan
    from pandas import read_csv, NA

    headers = {
        "User-Agent": "Mozilla/5.0 (Linux; U; Android 4.0.4; en-us; Glass 1 Build/IMM76L; XE16.2) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30"  # noqa  # pylint: disable=line-too-long
    }
    url = f"https://download.bls.gov/pub/time.series/{survey.lower()}/{survey.lower()}.{asset.lower()}"
    res = make_request(url=url, method="GET", headers=headers)

    if res.status_code != 200:
        return None

    df = read_csv(StringIO(res.text), sep="\t", low_memory=False, dtype="object")
    df.columns = [d.strip() for d in df.columns]
    df = df.apply(lambda x: x.str.strip() if x.dtype == "object" else x)

    return df.replace({"''": None, '""': None, NA: None, "nan": None, nan: None})


async def download_category_series_ids(category) -> tuple[list, dict]:
    """Download all series ids for a category of survey, along with the code maps.
    This should only be required for updating static files.
    """
    # pylint: disable=import-outside-toplevel
    from numpy import nan  # noqa
    from openbb_core.provider.utils.errors import EmptyDataError
    from openbb_bls.utils.constants import SURVEY_CATEGORY_MAP, SURVEY_NAMES

    series_ids: list = []
    series_codes: dict = {}

    if category not in SURVEY_CATEGORY_MAP:
        raise EmptyDataError(
            f"Category {category} is not a supported choice. Choose from {list(SURVEY_CATEGORY_MAP)}"
        )

    async def get_all_series_ids(survey):
        """Get an asset in the FTP download folder of the two-letter survey code."""
        if survey in ["ch", "cs", "fw", "is", "nw", "oe", "yy"]:
            return
        data = await get_survey_asset(survey, "series")
        for col in ["series_title", "survey_name"]:
            if col not in data.columns:  # type: ignore
                data.loc[:, col] = None  # type: ignore
        codes = [d for d in data.columns if "code" in d and "periodicity" not in d]  # type: ignore
        ids = data.get(["series_id", "series_title"] + codes).copy()  # type: ignore

        if ids is None or ids.empty:
            return

        ids = ids.rename(columns={"footnote_codes": "footnote_code"})
        ids = ids.astype(str).replace({"nan": None, "''": None, "": None})

        ids.loc[:, "survey_name"] = SURVEY_NAMES.get(survey.upper(), None)
        ids = ids[ids.series_id.astype(str).str.startswith(survey.upper())]

        if ids is None or ids.empty:
            return

        # Get the code maps for the survey and convert the codes in the series table.
        code_map: dict = {}
        new_codes = [d.split("_")[0] for d in codes]

        for code in new_codes:
            code = "datatype" if code == "data" else code  # noqa
            code_dict: dict = {}
            code_data = await get_survey_asset(survey, code)
            if code_data is None or code_data.empty:
                continue
            code = (  # noqa
                "data_type" if code == "datatype" and survey == "ce" else code
            )
            code_data = code_data.rename(
                columns={
                    col: f"{code}_name"
                    for col in code_data.columns
                    if col in [f"{code}_text", f"{code}_title"]
                }
            )
            if f"{code}_code" in code_data.columns:
                if (
                    code_data.index.dtype == "object"
                    and code_data.get(f"{code}_name").isnull().all()  # type: ignore
                ):
                    code_data.loc[:, f"{code}_name"] = code_data.loc[
                        :, f"{code}_code"
                    ].copy()
                    code_data.loc[:, f"{code}_code"] = code_data.index.copy()  # type: ignore
                    code_data = code_data.reset_index(drop=True)
                code_dict = (
                    code_data.set_index(f"{code}_code")[[f"{code}_name"]]
                    .to_dict()
                    .get(f"{code}_name", code_dict)
                )
            else:
                code_dict = code_data.to_dict(orient="series")

            code_map[f"{code}_code"] = code_dict
            ids[f"{code}_code"] = [code_dict.get(d, d) for d in ids[f"{code}_code"]]

        # Footnotes may be comma-separated, so we need to expand them.
        if "footnote_code" in ids.columns:
            expanded_data = []
            for item in ids["footnote_code"]:
                if (
                    item
                    and isinstance(item, str)
                    and "," in item
                    and any(char.isdigit() for char in item)
                ):
                    expanded_data.append(
                        " ".join(
                            [
                                code_map["footnote_code"].get(sub_item, sub_item)
                                for sub_item in item.split(",")
                                if code_map["footnote_code"].get(sub_item) is not None
                            ]
                        )
                    )
                else:
                    expanded_data.append(item)
            ids["footnote_code"] = expanded_data

        ids = ids.replace({nan: None, "nan": None, "''": None}).to_dict(
            orient="records"
        )
        series_ids.extend(ids)
        series_codes.update({survey: code_map})

    # Iterate over the all the surveys in the category and generate the tables and code maps.
    # The FTP doesn't seem to like a flood of requests, so we do this operation in series.
    for survey in SURVEY_CATEGORY_MAP[category]:
        await get_all_series_ids(survey.lower())

    return series_ids, series_codes


def open_asset(asset: str) -> Union["DataFrame", dict]:
    """Open a static file asset for series IDs or code maps."""
    # pylint: disable=import-outside-toplevel
    import os  # noqa
    import json
    from importlib.resources import files
    from pathlib import Path
    from numpy import nan
    from openbb_core.app.model.abstract.error import OpenBBError
    from pandas import read_csv

    if ".xz" not in asset and "series" in asset:
        asset = asset + ".xz"
    elif ".json" not in asset and "codes" in asset:
        asset = asset + ".json"
    elif ".json" in asset or ".xz" in asset:
        pass
    else:
        raise OpenBBError(f"Asset '{asset}' not supported. Expected .json or .xz file.")

    assets_path = Path(str(files("openbb_bls").joinpath("assets")))

    if not os.path.exists(assets_path.joinpath(asset)):
        raise OpenBBError(f"Asset '{asset}' not found.")

    if asset.endswith(".json"):
        with open(assets_path.joinpath(asset)) as f:
            return json.load(f)
    else:
        with open(assets_path.joinpath(asset), "rb") as f:
            df = read_csv(f, compression="xz", low_memory=False, dtype="str")
        return df.replace({nan: None, "nan": None, "''": None}).dropna(
            how="all", axis=1
        )


async def update_static_asset(category: str) -> None:
    """Update a static file assets with series IDs and code maps for a given category.
    Do not use unless the static files in the assets folder require updating.
    """
    # pylint: disable=import-outside-toplevel
    import json  # noqa
    from importlib.resources import files
    from pathlib import Path
    from openbb_core.app.model.abstract.error import OpenBBError
    from openbb_bls.utils.constants import SURVEY_CATEGORY_NAMES
    from numpy import nan
    from pandas import DataFrame

    if category not in SURVEY_CATEGORY_NAMES:
        raise OpenBBError(
            f"Category '{category}' not found. Choose from {list(SURVEY_CATEGORY_NAMES)}"
        )

    try:
        ids, codes = await download_category_series_ids(category)
    except Exception as e:  # pylint: disable=broad-except
        raise OpenBBError(f"Failed to download {category} -> {e}") from e

    assets_path = Path(str(files("openbb_bls").joinpath("assets")))

    # Save the code map to a JSON file.
    if codes:
        with open(assets_path.joinpath(f"{category}_codes.json"), "w") as f:
            json.dump(codes, f, indent=4)

    # Save the series IDs to a CSV file.
    if ids:
        df = DataFrame(ids)
        df = df.replace({nan: None, "nan": None, "''": None}).dropna(how="all", axis=1)
        df.to_csv(
            assets_path.joinpath(f"{category}_series.xz"), index=False, compression="xz"
        )



"""BLS Search Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_bls.utils.constants import SURVEY_CATEGORIES, SURVEY_CATEGORY_NAMES
from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.annotated_result import AnnotatedResult
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.bls_search import (
    SearchData,
    SearchQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class BlsSearchQueryParams(SearchQueryParams):
    """BLS Search Query Parameters."""

    __json_schema_extra__ = {
        "category": {
            "multiple_items_allowed": False,
            "choices": list(SURVEY_CATEGORY_NAMES),
        },
    }

    category: SURVEY_CATEGORIES = Field(
        description="""The category of BLS survey to search within.
        An empty search query will return all series within the category. Options are:
        \n    cpi - Consumer Price Index
        \n    pce - Personal Consumption Expenditure
        \n    ppi - Producer Price Index
        \n    ip - Industry Productivity
        \n    jolts - Job Openings and Labor Turnover Survey
        \n    nfp - Nonfarm Payrolls
        \n    cps - Current Population Survey
        \n    lfs - Labor Force Statistics
        \n    wages - Wages
        \n    ec - Employer Costs
        \n    sla - State and Local Area Employment
        \n    bed - Business Employment Dynamics
        \n    tu - Time Use
        """,
    )
    include_extras: bool = Field(
        default=False,
        description="Include additional information in the search results."
        + " Extra fields returned are metadata and vary by survey."
        + " Fields are undefined strings that typically have names ending with '_code'.",
    )
    include_code_map: bool = Field(
        default=False,
        description="When True, includes the complete code map for eaçh survey in the category,"
        + " returned separately as a nested JSON to the `extras['results_metadata']` property of the response."
        + " Example content is the NAICS industry map for PPI surveys."
        + " Each code is a value within the 'symbol' of the time series.",
    )


class BlsSearchData(SearchData):
    """BLS Search Data."""

    __alias_dict__ = {
        "symbol": "series_id",
        "title": "series_title",
    }


class BlsSearchFetcher(Fetcher[BlsSearchQueryParams, list[BlsSearchData]]):
    """BLS Search Fetcher."""

    require_credentials = False

    @staticmethod
    def transform_query(params: dict[str, Any]) -> BlsSearchQueryParams:
        """Transform query parameters."""
        return BlsSearchQueryParams(**params)

    @staticmethod
    def extract_data(
        query: BlsSearchQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Extract the data."""
        # pylint: disable=import-outside-toplevel
        from openbb_bls.utils.helpers import open_asset
        from pandas import Series

        try:
            df = open_asset(f"{query.category}_series")
        except OpenBBError as e:
            raise e from e

        terms = [term.strip() for term in query.query.split(";")] if query.query else []

        if not terms:
            records = (
                df.to_dict(orient="records")
                if query.include_extras is True
                else df.filter(
                    items=["series_id", "series_title", "survey_name"], axis=1
                ).to_dict(orient="records")
            )
        else:
            combined_mask = Series([True] * len(df))
            for term in terms:
                mask = df.apply(
                    lambda row, term=term: row.astype(str).str.contains(
                        term, case=False, regex=True, na=False
                    )
                ).any(axis=1)
                combined_mask &= mask

            matches = df[combined_mask]

            if matches.empty:
                raise EmptyDataError("No results found for the provided query.")

            records = (
                matches.to_dict(orient="records")
                if query.include_extras is True
                else matches.filter(
                    items=["series_id", "series_title", "survey_name"], axis=1
                ).to_dict(orient="records")
            )

        return records

    @staticmethod
    def transform_data(
        query: BlsSearchQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> AnnotatedResult[list[BlsSearchData]]:
        """Transform the data."""
        metadata: dict = {}
        if query.include_code_map is True:
            # pylint: disable=import-outside-toplevel
            from openbb_bls.utils.helpers import open_asset

            try:
                metadata = open_asset(f"{query.category}_codes")
            except OpenBBError as e:
                raise e from e

        return AnnotatedResult(
            result=[BlsSearchData.model_validate(d) for d in data],
            metadata=metadata,
        )



"""BLS Series Model."""

# pylint: disable=unused-argument

from typing import Any
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.annotated_result import AnnotatedResult
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.bls_series import (
    SeriesData,
    SeriesQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class BlsSeriesQueryParams(SeriesQueryParams):
    """BLS Series Query Parameters."""

    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
    }

    calculations: bool = Field(
        default=True,
        description="Include calculations in the response, if available. Default is True.",
    )
    annual_average: bool = Field(
        default=False,
        description="Include annual averages in the response, if available. Default is False.",
    )
    aspects: bool = Field(
        default=False,
        description="Include all aspects associated with a data point for a given BLS series ID, if available."
        + " Returned with the series metadata, under `extras` of the response object. Default is False.",
    )


class BlsSeriesData(SeriesData):
    """BLS Series Data."""

    change_1M: float | None = Field(
        default=None,
        description="One month change in value.",
    )
    change_3M: float | None = Field(
        default=None,
        description="Three month change in value.",
    )
    change_6M: float | None = Field(
        default=None,
        description="Six month change in value.",
    )
    change_12M: float | None = Field(
        default=None,
        description="One year change in value.",
    )
    change_percent_1M: float | None = Field(
        default=None,
        description="One month change in percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    change_percent_3M: float | None = Field(
        default=None,
        description="Three month change in percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    change_percent_6M: float | None = Field(
        default=None,
        description="Six month change in percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    change_percent_12M: float | None = Field(
        default=None,
        description="One year change in percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    latest: bool | None = Field(
        default=None,
        description="Latest value indicator.",
    )
    footnotes: str | None = Field(
        default=None,
        description="Footnotes accompanying the value.",
    )


class BlsSeriesFetcher(Fetcher[BlsSeriesQueryParams, list[BlsSeriesData]]):
    """BLS Series Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> BlsSeriesQueryParams:
        """Transform query parameters."""
        return BlsSeriesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: BlsSeriesQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> dict:
        """Extract the data from the BLS API."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        from datetime import datetime, timedelta
        from openbb_bls.utils.helpers import get_bls_timeseries

        api_key = credentials.get("bls_api_key") if credentials else ""
        symbols = (
            query.symbol.split(",") if isinstance(query.symbol, str) else query.symbol
        )
        now = datetime.now()
        start_year = (
            query.start_date.year
            if query.start_date
            else (now - timedelta(weeks=52 * 3)).year
        )
        end_year = query.end_date.year if query.end_date else now.year
        results: dict = {"data": [], "messages": [], "metadata": {}}
        messages: list = []

        # The max number of symbols per request is 50.
        # The max year range is 20.
        # We chunk the request to handle the provided start/end date by the user.
        def chunk_list(lst, chunk_size):
            """Yield successive chunks from lst of size chunk_size."""
            for i in range(0, len(lst), chunk_size):
                yield lst[i : i + chunk_size]

        def chunk_years(start_year, end_year, chunk_size):
            """Yield successive year ranges of size chunk_size."""
            for year in range(start_year, end_year + 1, chunk_size):
                yield (year, min(year + chunk_size - 1, end_year))

        # Define a function to wrap as a coroutine.
        async def make_query(symbol, start, end):
            """Make a query to the BLS API."""
            data = await get_bls_timeseries(
                api_key=api_key,
                series_ids=symbol,
                start_year=start,
                end_year=end,
                calculations=query.calculations,
                catalog=True,
                annual_average=query.annual_average,
                aspects=query.aspects,
            )
            if isinstance(data, dict):
                results.update(
                    {
                        "data": results.get("data", []) + data.get("data", []),
                        "messages": list(
                            set(results.get("messages", []) + data.get("messages", []))
                        ),
                        "metadata": {
                            **results.get("metadata", {}),
                            **data.get("metadata", {}),
                        },
                    }
                )
            elif isinstance(data, EmptyDataError) and data.message:
                messages.append(data.__dict__.get("message", ""))

        # Create a list of tasks to run based on the API query limitations.
        tasks: list = []

        for symbol_chunk in chunk_list(symbols, 50):
            for year_range in chunk_years(start_year, end_year, 20):
                tasks.append(
                    asyncio.create_task(
                        make_query(
                            symbol_chunk,
                            year_range[0],
                            year_range[1],
                        )
                    )
                )

        await asyncio.gather(*tasks)

        if not results.get("data"):
            if messages:
                raise OpenBBError(",".join(set(messages)))
            raise EmptyDataError("The request was returned empty.")

        return results

    @staticmethod
    def transform_data(
        query: BlsSeriesQueryParams,
        data: dict,
        **kwargs: Any,
    ) -> AnnotatedResult[list[BlsSeriesData]]:
        """Transform the data."""
        series_data = data.get("data", [])
        messages = data.get("messages", [])
        metadata = data.get("metadata", {})
        if messages:
            for message in messages:
                warn(message)

        results = sorted(
            [BlsSeriesData.model_validate(series) for series in series_data],
            key=lambda x: (x.date, x.symbol),
        )

        if query.start_date is not None:
            results = [r for r in results if r.date >= query.start_date]

        if query.end_date is not None:
            results = [r for r in results if r.date <= query.end_date]

        return AnnotatedResult(
            result=results,
            metadata=metadata,
        )



"""BLS Provider Constants."""

from typing import Literal

PPI_SURVEYS = ["WP", "PC"]
PCE_SURVEYS = ["CX"]
CPI_SURVEYS = ["AP", "CU", "CW", "LI", "SU", "EI"]
PRODUCTIVITY_SURVEYS = ["IP", "PR", "MP"]
LABOR_FORCE_STATISTICS = ["LN", "FM", "IN", "WS"]
CURRENT_POPULATION_SURVEY = ["LE", "LU"]
LOCAL_EMPLOYMENT_SURVEYS = ["LA", "SM"]
JOLTS_SURVEYS = ["JL", "JT"]
NFP_SURVEYS = ["CE"]
WAGES_SURVEYS = ["CI", "WM"]
EMPLOYER_COSTS = ["CM", "CC"]
BUSINESS_EMPLOYMENT_DYNAMICS = ["BD"]
TIME_USE_SURVEYS = ["TU"]

SUPPORTED_SURVEYS = sorted(
    list(
        PPI_SURVEYS
        + PCE_SURVEYS
        + CPI_SURVEYS
        + PRODUCTIVITY_SURVEYS
        + LABOR_FORCE_STATISTICS
        + LOCAL_EMPLOYMENT_SURVEYS
        + CURRENT_POPULATION_SURVEY
        + JOLTS_SURVEYS
        + NFP_SURVEYS
        + WAGES_SURVEYS
        + EMPLOYER_COSTS
        + BUSINESS_EMPLOYMENT_DYNAMICS
        + TIME_USE_SURVEYS
    )
)

SURVEY_CATEGORY_MAP = {
    "cpi": [d.lower() for d in CPI_SURVEYS],
    "pce": [d.lower() for d in PCE_SURVEYS],
    "ppi": [d.lower() for d in PPI_SURVEYS],
    "ip": [d.lower() for d in PRODUCTIVITY_SURVEYS],
    "jolts": [d.lower() for d in JOLTS_SURVEYS],
    "nfp": [d.lower() for d in NFP_SURVEYS],
    "cps": [d.lower() for d in CURRENT_POPULATION_SURVEY],
    "lfs": [d.lower() for d in LABOR_FORCE_STATISTICS],
    "wages": [d.lower() for d in WAGES_SURVEYS],
    "ec": [d.lower() for d in EMPLOYER_COSTS],
    "sla": [d.lower() for d in LOCAL_EMPLOYMENT_SURVEYS],
    "bed": [d.lower() for d in BUSINESS_EMPLOYMENT_DYNAMICS],
    "tu": [d.lower() for d in TIME_USE_SURVEYS],
}

SURVEY_CATEGORY_NAMES = {
    "cpi": "Consumer Price Index",
    "pce": "Personal Consumption Expenditure",
    "ppi": "Producer Price Index",
    "ip": "Industry Productivity",
    "jolts": "Job Openings and Labor Turnover Survey",
    "nfp": "Nonfarm Payrolls",
    "cps": "Current Population Survey",
    "lfs": "Labor Force Statistics",
    "wages": "Wages",
    "ec": "Employer Costs",
    "sla": "State and Local Area Employment",
    "bed": "Business Employment Dynamics",
    "tu": "Time Use",
}

SURVEY_CATEGORIES = Literal[
    "cpi",
    "pce",
    "ppi",
    "ip",
    "jolts",
    "nfp",
    "cps",
    "lfs",
    "wages",
    "ec",
    "sla",
    "bed",
    "tu",
]

ALL_SURVEYS = [
    {
        "survey_name": "Consumer Price Index - Average Price Data",
        "survey_abbreviation": "AP",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Business Employment Dynamics",
        "survey_abbreviation": "BD",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Collective Bargaining Agreements-State and Local Government",
        "survey_abbreviation": "BG",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Collective Bargaining Agreements-Private Sector",
        "survey_abbreviation": "BP",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Biennial Nonfatal Case and Demographic numbers and rates: selected characteristics",
        "survey_abbreviation": "CB",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Employer Costs for Employee Compensation",
        "survey_abbreviation": "CC",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Nonfatal cases involving days away from work: selected characteristics",
        "survey_abbreviation": "CD",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Employment, Hours, and Earnings from the Current Employment Statistics survey (National)",
        "survey_abbreviation": "CE",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Census of Fatal Occupational Injuries",
        "survey_abbreviation": "CF",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Nonfatal cases involving days away from work: selected characteristics (2003 - 2010)",
        "survey_abbreviation": "CH",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Employment Cost Index",
        "survey_abbreviation": "CI",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Employer Costs for Employee Compensation",
        "survey_abbreviation": "CM",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Nonfatal cases involving days away from work: selected characteristics (2011 forward)",
        "survey_abbreviation": "CS",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Consumer Price Index - All Urban Consumers",
        "survey_abbreviation": "CU",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Consumer Price Index - Urban Wage Earners and Clerical Workers",
        "survey_abbreviation": "CW",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Consumer Expenditure Survey",
        "survey_abbreviation": "CX",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Import/Export Price Indexes",
        "survey_abbreviation": "EI",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Quarterly Census of Employment and Wages (SIC)",
        "survey_abbreviation": "EW",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Census of Fatal Occupational Injuries (2003 - 2010)",
        "survey_abbreviation": "FI",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Marital and family labor force statistics from the Current Population Survey",
        "survey_abbreviation": "FM",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Census of Fatal Occupational Injuries (2011 forward)",
        "survey_abbreviation": "FW",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Nonfatal cases involving days away from work: Selected Characteristics (2002)",
        "survey_abbreviation": "HC",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Occupational injuries and illnesses: industry data (pre-1989)",
        "survey_abbreviation": "HS",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Occupational injuries and illnesses: industry data",
        "survey_abbreviation": "II",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "International Labor Comparison",
        "survey_abbreviation": "IN",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Industry Productivity",
        "survey_abbreviation": "IP",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Occupational injuries and illnesses industry data",
        "survey_abbreviation": "IS",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Job Openings and Labor Turnover Survey",
        "survey_abbreviation": "JL",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Job Openings and Labor Turnover Survey",
        "survey_abbreviation": "JT",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Local Area Unemployment Statistics",
        "survey_abbreviation": "LA",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Weekly and hourly earnings data from the Current Population Survey",
        "survey_abbreviation": "LE",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Consumer Price Index - Department Store Inventory Price Index",
        "survey_abbreviation": "LI",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Labor Force Statistics from the Current Population Survey",
        "survey_abbreviation": "LN",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Union affiliation data from the Current Population Survey",
        "survey_abbreviation": "LU",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Major Sector Total Factor Productivity",
        "survey_abbreviation": "MP",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "National Compensation Survey-Benefits",
        "survey_abbreviation": "NB",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "National Compensation Survey",
        "survey_abbreviation": "NC",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "National Compensation Survey",
        "survey_abbreviation": "NW",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Occupational Employment and Wage Statistics",
        "survey_abbreviation": "OE",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Occupational Requirements",
        "survey_abbreviation": "OR",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Producer Price Index Industry Data",
        "survey_abbreviation": "PC",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Major Sector Productivity and Costs",
        "survey_abbreviation": "PR",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Occupational injuries and illnesses: industry data (1989-2001)",
        "survey_abbreviation": "SH",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Occupational injuries and illnesses: industry data (2002)",
        "survey_abbreviation": "SI",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "State and Area Employment, Hours, and Earnings",
        "survey_abbreviation": "SM",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Consumer Price Index - Chained Consumer Price Index",
        "survey_abbreviation": "SU",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "American Time Use",
        "survey_abbreviation": "TU",
        "allowsNetChange": "false",
        "allowsPercentChange": "false",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Wage Modeling",
        "survey_abbreviation": "WM",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "false",
    },
    {
        "survey_name": "Producer Price Index-Commodities",
        "survey_abbreviation": "WP",
        "allowsNetChange": "false",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
    {
        "survey_name": "Work Stoppage Data",
        "survey_abbreviation": "WS",
        "allowsNetChange": "true",
        "allowsPercentChange": "true",
        "hasAnnualAverages": "true",
    },
]

SURVEY_NAMES = {
    d["survey_abbreviation"].lower(): d["survey_name"]
    for d in ALL_SURVEYS
    if d["survey_abbreviation"] in SUPPORTED_SURVEYS
}
